#!/usr/bin/env node
"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const tslib_1 = require("tslib");
const fs_1 = require("fs");
const path_1 = require("path");
const util_1 = require("util");
// Load Environment before BlobServerFactory to make sure args works properly
const Environment_1 = tslib_1.__importDefault(require("./common/Environment"));
// tslint:disable-next-line:ordered-imports
const BlobServerFactory_1 = require("./blob/BlobServerFactory");
const Logger = tslib_1.__importStar(require("./common/Logger"));
const QueueConfiguration_1 = tslib_1.__importDefault(require("./queue/QueueConfiguration"));
const QueueServer_1 = tslib_1.__importDefault(require("./queue/QueueServer"));
const constants_1 = require("./queue/utils/constants");
// tslint:disable:no-console
const accessAsync = util_1.promisify(fs_1.access);
/**
 * Entry for Azurite services.
 */
async function main() {
    // Initialize and validate environment values from command line parameters
    const env = new Environment_1.default();
    const location = await env.location();
    await accessAsync(location);
    const debugFilePath = await env.debug();
    if (debugFilePath !== undefined) {
        await accessAsync(path_1.dirname(debugFilePath));
    }
    const blobServerFactory = new BlobServerFactory_1.BlobServerFactory();
    const blobServer = await blobServerFactory.createServer(env);
    const blobConfig = blobServer.config;
    // TODO: Align with blob DEFAULT_BLOB_PERSISTENCE_ARRAY
    // TODO: Join for all paths in the array
    constants_1.DEFAULT_QUEUE_PERSISTENCE_ARRAY[0].locationPath = path_1.join(location, constants_1.DEFAULT_QUEUE_PERSISTENCE_PATH);
    const queueConfig = new QueueConfiguration_1.default(env.queueHost(), env.queuePort(), path_1.join(location, constants_1.DEFAULT_QUEUE_LOKI_DB_PATH), path_1.join(location, constants_1.DEFAULT_QUEUE_EXTENT_LOKI_DB_PATH), constants_1.DEFAULT_QUEUE_PERSISTENCE_ARRAY, !env.silent(), undefined, env.debug() !== undefined, await env.debug(), env.loose());
    // We use logger singleton as global debugger logger to track detailed outputs cross layers
    // Note that, debug log is different from access log which is only available in request handler layer to
    // track every request. Access log is not singleton, and initialized in specific RequestHandlerFactory implementations
    // Enable debug log by default before first release for debugging purpose
    Logger.configLogger(blobConfig.enableDebugLog, blobConfig.debugLogFilePath);
    // Create server instance
    const queueServer = new QueueServer_1.default(queueConfig);
    // Start server
    console.log(`Azurite Blob service is starting on ${blobConfig.host}:${blobConfig.port}`);
    await blobServer.start();
    console.log(`Azurite Blob service successfully listens on ${blobServer.getHttpServerAddress()}`);
    // Start server
    console.log(`Azurite Queue service is starting on ${queueConfig.host}:${queueConfig.port}`);
    await queueServer.start();
    console.log(`Azurite Queue service successfully listens on ${queueServer.getHttpServerAddress()}`);
    // Handle close event
    const blobBeforeCloseMessage = `Azurite Blob service is closing...`;
    const blobAfterCloseMessage = `Azurite Blob service successfully closed`;
    const queueBeforeCloseMessage = `Azurite Queue service is closing...`;
    const queueAfterCloseMessage = `Azurite Queue service successfully closed`;
    process
        .once("message", msg => {
        if (msg === "shutdown") {
            console.log(blobBeforeCloseMessage);
            blobServer.close().then(() => {
                console.log(blobAfterCloseMessage);
            });
            console.log(queueBeforeCloseMessage);
            queueServer.close().then(() => {
                console.log(queueAfterCloseMessage);
            });
        }
    })
        .once("SIGINT", () => {
        console.log(blobBeforeCloseMessage);
        blobServer.close().then(() => {
            console.log(blobAfterCloseMessage);
        });
        console.log(queueBeforeCloseMessage);
        queueServer.close().then(() => {
            console.log(queueAfterCloseMessage);
        });
    });
}
main().catch(err => {
    console.error(`Exit due to unhandled error: ${err.message}`);
    process.exit(1);
});
//# sourceMappingURL=azurite.js.map