"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const tslib_1 = require("tslib");
const v4_1 = tslib_1.__importDefault(require("uuid/v4"));
const Logger_1 = tslib_1.__importDefault(require("../../common/Logger"));
const utils_1 = require("../../common/utils/utils");
const QueueStorageContext_1 = tslib_1.__importDefault(require("../context/QueueStorageContext"));
const StorageErrorFactory_1 = tslib_1.__importDefault(require("../errors/StorageErrorFactory"));
const constants_1 = require("../utils/constants");
const utils_2 = require("../utils/utils");
/**
 * A middleware extract related queue service context.
 *
 * @export
 * @param {Request} req An express compatible Request object
 * @param {Response} res An express compatible Response object
 * @param {NextFunction} next An express middleware next callback
 */
function queueStorageContextMiddleware(req, res, next) {
    // Set server header in every Azurite response
    res.setHeader(constants_1.HeaderConstants.SERVER, `Azurite-Queue/${constants_1.VERSION}`);
    const requestID = v4_1.default();
    const apiVersion = req.header(constants_1.HeaderConstants.X_MS_VERSION);
    if (apiVersion !== undefined) {
        utils_1.checkApiVersion(apiVersion, constants_1.ValidAPIVersions, requestID);
    }
    const queueContext = new QueueStorageContext_1.default(res.locals, constants_1.DEFAULT_QUEUE_CONTEXT_PATH);
    queueContext.startTime = new Date();
    queueContext.xMsRequestID = requestID;
    Logger_1.default.info(`QueueStorageContextMiddleware: RequestMethod=${req.method} RequestURL=${req.protocol}://${req.hostname}${req.url} RequestHeaders:${JSON.stringify(req.headers)} ClientIP=${req.ip} Protocol=${req.protocol} HTTPVersion=${req.httpVersion}`, requestID);
    const [account, queue, message, messageId, isSecondary] = extractStoragePartsFromPath(req.path);
    queueContext.account = account;
    queueContext.queue = queue;
    queueContext.message = message;
    queueContext.messageId = messageId;
    queueContext.isSecondary = isSecondary;
    // Emulator's URL pattern is like http://hostname:port/account/queue/messages
    // Create a router to exclude account name from req.path, as url path in swagger doesn't include account
    // Exclude account name from req.path for dispatchMiddleware
    queueContext.dispatchPattern =
        queue !== undefined
            ? message !== undefined
                ? messageId !== undefined
                    ? `/queue/messages/messageId`
                    : `/queue/messages`
                : `/queue`
            : "/";
    // The value of queue may be "" in some cases, e.g. list queue with .../accountname/?comp=list...
    if (req.query &&
        (req.query.restype === "service" || req.query.comp === "list")) {
        queueContext.dispatchPattern = "/";
    }
    queueContext.authenticationPath = req.path;
    if (isSecondary) {
        const pos = queueContext.authenticationPath.search(constants_1.SECONDARY_SUFFIX);
        queueContext.authenticationPath =
            queueContext.authenticationPath.substr(0, pos) +
                queueContext.authenticationPath.substr(pos + constants_1.SECONDARY_SUFFIX.length);
    }
    if (account === undefined) {
        const handlerError = StorageErrorFactory_1.default.getInvalidQueryParameterValue(requestID);
        Logger_1.default.error(`QueueStorageContextMiddleware: QueueStorageContextMiddleware: ${handlerError.message}`, requestID);
        return next(handlerError);
    }
    // If it is not and account operations, then the queue name should be validated.
    if (queueContext.dispatchPattern !== "/" && queue !== undefined) {
        const nameValidateStatus = utils_2.isValidName(queue);
        if (nameValidateStatus === utils_2.nameValidateCode.invalidUri) {
            const handlerError = StorageErrorFactory_1.default.getInvalidUri(requestID, {
                UriPath: `/${queue}`
            });
            return next(handlerError);
        }
        if (nameValidateStatus === utils_2.nameValidateCode.outOfRange) {
            const handlerError = StorageErrorFactory_1.default.getOutOfRangeName(requestID);
            return next(handlerError);
        }
        if (nameValidateStatus === utils_2.nameValidateCode.invalidName) {
            const handlerError = StorageErrorFactory_1.default.getInvalidResourceName(requestID);
            return next(handlerError);
        }
    }
    Logger_1.default.info(`QueueStorageContextMiddleware: Account=${account} Queue=${queue} Message=${message} MessageId=${messageId}`, requestID);
    next();
}
exports.default = queueStorageContextMiddleware;
/**
 * Extract storage account, queue, and massages from URL path.
 *
 * @param {string} path
 * @returns {([string | undefined, string | undefined, string | undefined, boolean | undefined])}
 */
function extractStoragePartsFromPath(path) {
    let account;
    let queue;
    let message;
    let messageId;
    let isSecondary = false;
    const decodedPath = decodeURIComponent(path);
    const normalizedPath = decodedPath.startsWith("/")
        ? decodedPath.substr(1)
        : decodedPath; // Remove starting "/"
    const parts = normalizedPath.split("/");
    account = parts[0];
    queue = parts[1];
    // For delete and update, it is messages/messageid?popreceipt=string-value
    message = parts[2];
    messageId = parts[3];
    if (account.endsWith(constants_1.SECONDARY_SUFFIX)) {
        account = account.substr(0, account.length - constants_1.SECONDARY_SUFFIX.length);
        isSecondary = true;
    }
    return [account, queue, message, messageId, isSecondary];
}
exports.extractStoragePartsFromPath = extractStoragePartsFromPath;
//# sourceMappingURL=queueStorageContext.middleware.js.map