/// <reference types="node" />
import * as http from "http";
import * as https from "https";
import ConfigurationBase from "./ConfigurationBase";
import ICleaner from "./ICleaner";
import IRequestListenerFactory from "./IRequestListenerFactory";
export declare type RequestListener = (request: http.IncomingMessage, response: http.ServerResponse) => void;
export declare enum ServerStatus {
    Closed = "Closed",
    Starting = "Starting",
    Running = "Running",
    Closing = "Closing"
}
/**
 * Abstract Server class for Azurite HTTP or HTTPS Servers.
 *
 * @export
 * @abstract
 * @class Server
 */
export default abstract class ServerBase implements ICleaner {
    readonly host: string;
    readonly port: number;
    readonly httpServer: http.Server | https.Server;
    readonly config: ConfigurationBase;
    protected status: ServerStatus;
    /**
     * Creates an instance of HTTP or HTTPS server.
     *
     * @param {string} host Server host,for example, "127.0.0.1"
     * @param {number} port Server port, for example, 10000
     * @param {http.Server | https.Server} httpServer A HTTP or HTTPS server instance without request listener bound
     * @param {IRequestListenerFactory} requestListenerFactory A request listener factory
     * @memberof ServerBase
     */
    constructor(host: string, port: number, httpServer: http.Server | https.Server, requestListenerFactory: IRequestListenerFactory, config: ConfigurationBase);
    /**
     * Get HTTP server listening address and port string.
     * Note this may be different from host and port parameters values, because
     * when port is 0, system will select a rand port number for listening.
     * This method will return the port and address being used.
     *
     * @returns {string}
     * @memberof ServerBase
     */
    getHttpServerAddress(): string;
    getStatus(): ServerStatus;
    /**
     * Initialize and start the server to server incoming HTTP requests.
     * beforeStart() and afterStart() will be executed before and after start().
     *
     * @abstract
     * @returns {Promise<void>}
     * @memberof Server
     */
    start(): Promise<void>;
    /**
     * Dispose HTTP server and clean up other resources.
     *
     * beforeClose() and afterClose() will be executed before and after close().
     *
     * We name this method as close instead of dispose, because in practices, usually we cannot re-open the resources
     * disposed, but can re-open the resources closed.
     *
     * @abstract
     * @returns {Promise<void>}
     * @memberof Server
     */
    close(): Promise<void>;
    clean(): Promise<void>;
    /**
     * Async task before server starts.
     *
     * @protected
     * @abstract
     * @returns {Promise<void>}
     * @memberof Server
     */
    protected beforeStart(): Promise<void>;
    /**
     * Async task after server starts.
     *
     * @protected
     * @abstract
     * @returns {Promise<void>}
     * @memberof Server
     */
    protected afterStart(): Promise<void>;
    /**
     * Async task before server closes.
     *
     * @protected
     * @abstract
     * @returns {Promise<void>}
     * @memberof Server
     */
    protected beforeClose(): Promise<void>;
    /**
     * Async task after server closes.
     *
     * @protected
     * @abstract
     * @returns {Promise<void>}
     * @memberof Server
     */
    protected afterClose(): Promise<void>;
}
//# sourceMappingURL=ServerBase.d.ts.map