/// <reference types="node" />
import ILogger from "../ILogger";
import IExtentMetadataStore from "./IExtentMetadataStore";
import IExtentStore, { IExtentChunk, StoreDestinationArray } from "./IExtentStore";
/**
 * Persistency layer data store source implementation interacting with the storage media.
 * It provides the methods to read and write data with the storage.
 *
 * @export
 * @class FSExtentStore
 * @implements {IExtentStore}
 */
export default class FSExtentStore implements IExtentStore {
    private readonly persistencyConfiguration;
    private readonly logger;
    private readonly metadataStore;
    private readonly appendQueue;
    private readonly readQueue;
    private initialized;
    private closed;
    private activeWriteExtents;
    private activeWriteExtentsNumber;
    private persistencyPath;
    constructor(metadata: IExtentMetadataStore, persistencyConfiguration: StoreDestinationArray, logger: ILogger);
    isInitialized(): boolean;
    isClosed(): boolean;
    init(): Promise<void>;
    close(): Promise<void>;
    clean(): Promise<void>;
    /**
     * This method may create a new extent or append data to an existing extent.
     * Return the extent chunk information including the extentId, offset and count.
     *
     * @param {(NodeJS.ReadableStream | Buffer)} data
     * @param {string} [contextId]
     * @returns {Promise<IExtentChunk>}
     * @memberof FSExtentStore
     */
    appendExtent(data: NodeJS.ReadableStream | Buffer, contextId?: string): Promise<IExtentChunk>;
    /**
     * Read data from persistency layer according to the given IExtentChunk.
     *
     * @param {IExtentChunk} [extentChunk]
     * @returns {Promise<NodeJS.ReadableStream>}
     * @memberof FSExtentStore
     */
    readExtent(extentChunk?: IExtentChunk, contextId?: string): Promise<NodeJS.ReadableStream>;
    /**
     * Merge several extent chunks to a ReadableStream according to the offset and count.
     *
     * @param {(IExtentChunk)[]} extentChunkArray
     * @param {number} [offset=0]
     * @param {number} [count=Infinity]
     * @param {string} [contextId]
     * @returns {Promise<NodeJS.ReadableStream>}
     * @memberof FSExtentStore
     */
    readExtents(extentChunkArray: (IExtentChunk)[], offset?: number, count?: number, contextId?: string): Promise<NodeJS.ReadableStream>;
    /**
     * Delete the extents from persistency layer.
     *
     * @param {Iterable<string>} persistency
     * @returns {Promise<number>} Number of extents deleted
     * @memberof IExtentStore
     */
    deleteExtents(extents: Iterable<string>): Promise<number>;
    /**
     * Return its metadata store.
     *
     * @returns {IExtentMetadataStore}
     * @memberof IExtentStore
     */
    getMetadataStore(): IExtentMetadataStore;
    private streamPipe;
    /**
     * Check an extent is one of active extents or not.
     *
     * @private
     * @param {string} id
     * @returns {boolean}
     * @memberof FSExtentStore
     */
    private isActiveExtent;
    /**
     * Create a new append extent model for a new write directory.
     *
     * @private
     * @param {string} persistencyPath
     * @returns {IAppendExtent}
     * @memberof FSExtentStore
     */
    private createAppendExtent;
    /**
     * Select a new extent to append for an exist write directory.
     *
     * @private
     * @param {IAppendExtent} appendExtent
     * @memberof FSExtentStore
     */
    private getNewExtent;
    /**
     * Generate the file path for a new extent.
     *
     * @private
     * @param {string} extentId
     * @returns {string}
     * @memberof FSExtentStore
     */
    private generateExtentPath;
}
//# sourceMappingURL=FSExtentStore.d.ts.map